#!/usr/bin/env python
import os
import click
import random

import sys
import subprocess
import pkg_resources

import shutil

sachima_version = pkg_resources.require("sachima")[0].version
sys.path.insert(0, os.getcwd())

COLORS = {
    "black": "\u001b[30;1m",
    "red": "\u001b[31;1m",
    "green": "\u001b[32m",
    "yellow": "\u001b[33;1m",
    "blue": "\u001b[34;1m",
    "magenta": "\u001b[35m",
    "cyan": "\u001b[36m",
    "white": "\u001b[37m",
    # "yellow-background":"\u001b[43m",
    # "black-background":"\u001b[40m",
    # "cyan-background":"\u001b[46;1m",
}

CURRENT_DIR = os.getcwd()
WAFFLE_DIR = os.path.join(CURRENT_DIR, "..", "Waffle")


@click.group()
def sachima():
    logo = """
*********************************************************
                _     _                 
               | |   (_)                
 ___  __ _  ___| |__  _ _ __ ___   __ _ 
/ __|/ _` |/ __| '_ \| | '_ ` _ \ / _` |
\__ \ (_| | (__| | | | | | | | | | (_| |
|___/\__,_|\___|_| |_|_|_| |_| |_|\__,_|

Better Data Analysis                  version {} 
*********************************************************
    """.format(
        sachima_version
    )
    click.echo(random.choice(list(COLORS.values())) + logo)


def is_in_sachima_project():
    if os.path.isfile(os.path.join(os.getcwd(), "./sachima_config.py")):
        return True

    click.echo(
        "You should cd into your sachima project. Maybe you want to create your project first by running sachima init"
    )
    return False


@click.command(help="Print sachima version")
def version():
    click.echo(sachima_version)


@click.command(help="Get sachima middleware from github : get DessertsLab/pivot_table")
@click.option(
    "--path", default=os.path.join(os.getcwd(), "middleware"), help="project path",
)
@click.argument("middleware_name")
def get(path, middleware_name):
    if not is_in_sachima_project():
        return
    if not os.path.exists(path):
        os.makedirs(path)
    click.echo(
        "Getting middleware from https://github.com/{} and save to {}".format(
            middleware_name, path
        )
    )
    os.system(
        "git clone https://github.com/{} {}".format(
            middleware_name, os.path.join(path, middleware_name)
        )
    )


@click.command(help="Init a sachima project")
@click.argument("name")
def init(name):
    sachima_example_path = os.path.join(os.path.dirname(__file__), "init_example")
    current_init_path = os.path.join(os.getcwd(), name)
    click.echo("copying {} to {} ...".format(sachima_example_path, current_init_path))
    shutil.copytree(sachima_example_path, current_init_path)
    click.echo("Init a sachima project")


@click.command(help="Start sachima server")
def start():
    click.echo("start sachima server")


def sync_waffle():
    """
    DessertsLab/Waffle is frontend for sachima dev env
    pre download Waffle from github to the parent dir
    or update Waffle 
    """
    if not os.path.exists(WAFFLE_DIR):
        click.echo("Cloneing  DessertsLab/Waffle...")
        os.system(
            "git clone https://github.com/DessertsLab/Waffle.git {}".format(WAFFLE_DIR)
        )
        click.echo("Installing  DessertsLab/Waffle...")

        # os.system("cd {}".format(WAFFLE_DIR))
        os.chdir(WAFFLE_DIR)
        os.system("npm install")
        # os.system("cd {}".format(CURRENT_DIR))
        # os.system("npm install --prefix {}".format(WAFFLE_DIR))
    else:
        click.echo("Pulling  DessertsLab/Waffle...")
        os.system("git -C {0} pull origin master".format(WAFFLE_DIR))
        click.echo("Installing  DessertsLab/Waffle...")
        os.chdir(WAFFLE_DIR)
        os.system("npm install")
        # click.echo("cd {}".format(CURRENT_DIR))
        # os.system("cd {}".format(CURRENT_DIR))
        # os.system("npm install --prefix {}".format(WAFFLE_DIR))


def start_sachima():
    from sachima.sachima_http_server_flask import app

    sys.path.insert(0, os.getcwd())
    sys.dont_write_bytecode = True
    app.run(host="0.0.0.0", port=80, debug=False)


@click.command(help="Update sachima")
def update():
    click.echo("Updating sachima...")
    os.system("pip3 install -U sachima")
    click.echo("Updating waffle...")
    sync_waffle()


@click.command(help="Run sachima dev server with Waffle")
def run():
    if not is_in_sachima_project():
        return
    # sync_waffle()
    print("-" * 80)
    print(WAFFLE_DIR)
    print("-" * 80)
    # --prefix means start npm in a different directory
    cmd_line = "npm start --prefix {0}".format(WAFFLE_DIR)
    w = subprocess.Popen(
        cmd_line, shell=True, stdout=subprocess.PIPE, stderr=subprocess.STDOUT
    )
    s = subprocess.Popen(
        start_sachima(), shell=True, stdout=subprocess.PIPE, stderr=subprocess.STDOUT,
    )
    w.wait()
    s.wait()


sachima.add_command(get)
sachima.add_command(init)
sachima.add_command(start)
sachima.add_command(run)
sachima.add_command(version)
sachima.add_command(update)


if __name__ == "__main__":
    sachima()


# TODO: build Waffle and put it in sachima use this script start it
# import sys
# import thread
# import webbrowser
# import time

# import BaseHTTPServer, SimpleHTTPServer

# def start_server():
#     httpd = BaseHTTPServer.HTTPServer(('127.0.0.1', 3600), SimpleHTTPServer.SimpleHTTPRequestHandler)
#     httpd.serve_forever()

# thread.start_new_thread(start_server,())
# url = 'http://127.0.0.1:3600'
# webbrowser.open_new(url)

# while True:
#     try:
#         time.sleep(1)
#     except KeyboardInterrupt:
#         sys.exit(0)



import os
import requests
import json
import time

from sachima.log import logger
from sachima import conf

from io import BytesIO

BAIDU_GEO_TOKEN = conf.get("BAIDU_GEO_TOKEN")
QQ_GEO_TOKEN = conf.get("QQ_GEO_TOKEN")
AMAP_GEO_TOKEN = conf.get("AMAP_GEO_TOKEN")


def poi_compound_dict(
    dim={"K001": "subway", "K002": "hospital", "K003": "school"},
    dis=[500, 1000],
    lat=31.191869,
    lng=121.446756,
    onlycnt=False,
    onlyfarest=True,
    savetofile=False,
):
    """
    DIM = {
        "K001": "subway",
        "K002": "hospital",
        "K003": "school"
    }

    DIS = [500, 1000, 1500, 2000]

    lat = 31.191869 
    lng = 121.446756
    """
    res = {}
    for k in dim:
        for d in dis:
            key = k + "_" + str(d)  # eg. K001_1000
            key_cnt = key + "_CNT"  # eg. K001_1000_CNT
            poi_json = poi(lat, lng, dim.get(k), radius=d)
            if onlycnt is False:
                if onlyfarest is False:
                    res[key] = str(poi_json)
                elif d == max(dis):
                    res[key] = str(poi_json)
            res[key_cnt] = poi_json.get("count")

    return res


def poi(lat, lng, keywords, radius=1000):
    """
    Get poi information from amap \n
    https://lbs.amap.com/api/webservice/guide/api/search \n
    """
    logger.info((lat, lng, keywords, radius))
    if AMAP_GEO_TOKEN is None:
        logger.info("error: Must config AMAP_GEO_TOKEN in sachima_config.py")
        raise "Must config AMAP_GEO_TOKEN in sachima_config.py"

    url = "https://restapi.amap.com/v3/place/around"
    values = {
        "key": AMAP_GEO_TOKEN,
        "location": "{},{}".format(lng, lat),
        "keywords": keywords,
        "types": "",
        "radius": radius,  # default one mile
    }
    try:
        r = requests.get(url, values).json()
        logger.debug(r)
        if r.get("info") == 'USER_DAILY_QUERY_OVER_LIMIT':
            raise 'USER_DAILY_QUERY_OVER_LIMIT'
        return r
    except Exception as e:
        raise e


def panorama(lat, lng):
    """
    Panorama information query
    return Image Buffer
    you can open the buffer use PIL.Image.open(buffer)
    """
    if BAIDU_GEO_TOKEN is None:
        logger.info("error: Must config BAIDU_GEO_TOKEN in sachima_config.py")
        raise "Must config BAIDU_GEO_TOKEN in sachima_config.py"

    url = "http://api.map.baidu.com/panorama/v2"
    values = {"ak": BAIDU_GEO_TOKEN, "fov": 180, "location": "{},{}".format(lng, lat)}
    try:
        logger.info("Fetching {},{} panorama picture ...".format(lat, lng))
        r = requests.get(url, values).content
        b = BytesIO()
        b.write(r)
        return b
    except Exception as e:
        raise e


def fetchBaiduLatLng(address):
    """
    Use baidu map api to get the latitude and longitude of city from the Internet. \n
    http://lbsyun.baidu.com/index.php?title=webapi \n
    """
    if BAIDU_GEO_TOKEN is None:
        logger.info("error: Must config BAIDU_GEO_TOKEN in sachima_config.py")
        raise "Must config BAIDU_GEO_TOKEN in sachima_config.py"
    values = {
        "address": address,
        "ret_coordtype": "",
        "ak": BAIDU_GEO_TOKEN,
        "sn": "",
        "precise": 1,
        "output": "json",
        "callback": "",
    }
    url = "http://api.map.baidu.com/geocoder/v2/"
    try:
        r = requests.get(url, params=values).json()
        logger.info(r)
        if r.get("status") != 0:
            return {}
        else:
            return {
                "baidu_lat": r.get("result").get("location").get("lat"),
                "baidu_lng": r.get("result").get("location").get("lng"),
                "baidu_precise": r.get("result").get("precise"),
                "baidu_confidence": r.get("result").get("confidence"),
                "baidu_comprehension": r.get("result").get("comprehension"),
                "baidu_level": r.get("result").get("level"),
                "baidu_status": r.get("status"),
            }
    except Exception as e:
        logger.info("fetchBaiduGeo for %s error: %s" % (address, e))
        raise e


def fetchQQLatLng(address):
    """
    Use qq map api to get latitude and longitude from the Internet.
    https://lbs.qq.com/miniProgram/jsSdk/jsSdkGuide/methodGeocoder
    """
    if QQ_GEO_TOKEN is None:
        logger.info("error: Must config QQ_GEO_TOKEN in sachima_config.py")
        raise "Must config QQ_GEO_TOKEN in sachima_config.py"

    values = {"key": QQ_GEO_TOKEN, "output": "json", "address": address}

    url = "https://apis.map.qq.com/ws/geocoder/v1/"
    try:
        r = requests.get(url, params=values).json()
        if r.get("status") == 347:
            return {}
        else:
            return {
                "qq_lat": r.get("result").get("location").get("lat"),
                "qq_lng": r.get("result").get("location").get("lng"),
                "qq_title": r.get("result").get("title"),
                "qq_adcode": r.get("result").get("ad_info").get("adcode"),
                "qq_province": r.get("result")
                .get("address_components")
                .get("province"),
                "qq_city": r.get("result").get("address_components").get("city"),
                "qq_district": r.get("result")
                .get("address_components")
                .get("district"),
                "qq_street": r.get("result").get("address_components").get("street"),
                "qq_street_number": r.get("result")
                .get("address_components")
                .get("street_number"),
                "qq_similarity": r.get("result").get("similarity"),
                "qq_deviation": r.get("result").get("deviation"),
                "qq_reliability": r.get("result").get("reliability"),
                "qq_level": r.get("result").get("level"),
                "qq_status": r.get("status"),
                "qq_message": r.get("message"),
            }
    except Exception as e:
        logger.info("fetchQQGeo for %s error: %s！" % (address, e))
        raise e


def fetchAmapLatLng(address):
    """
    Use amap api to get latitude and longitude from the Internet.
    https://lbs.qq.com/miniProgram/jsSdk/jsSdkGuide/methodGeocoder
    """
    logger.info("fetching {} amap geo info".format(address))
    if AMAP_GEO_TOKEN is None:
        logger.info("error: Must config AMAP_GEO_TOKEN in sachima_config.py")
        raise "Must config AMAP_GEO_TOKEN in sachima_config.py"
    par = {"address": address, "key": AMAP_GEO_TOKEN}
    base = "http://restapi.amap.com/v3/geocode/geo"
    r = requests.get(base, par).json()
    geocodes = {}
    GPS = [None, None]
    if r.get("count") != "0":
        geocodes = r.get("geocodes")[0]
        GPS = geocodes.get("location").split(",")
    return {
        "amap_lat": GPS[1],
        "amap_lng": GPS[0],
        "amap_status": r.get("status"),
        "amap_info": r.get("info"),
        "amap_infocode": r.get("infocode"),
        "amap_count": r.get("count"),
        "amap_formatted_address": geocodes.get("formatted_address"),
        "amap_country": geocodes.get("country"),
        "amap_province": geocodes.get("province"),
        "amap_citycode": geocodes.get("citycode"),
        "amap_city": geocodes.get("city"),
        "amap_adcode": geocodes.get("adcode"),
        "amap_street": str(geocodes.get("street")),
        "amap_number": str(geocodes.get("number")),
        "amap_level": geocodes.get("level"),
    }



import logging
import logging.config
import sys
from sachima import conf
from tqdm import tqdm


class DummyTqdmFile(object):
    """ Dummy file-like that will write to tqdm
    https://github.com/tqdm/tqdm/issues/313
    """

    file = None

    def __init__(self, file):
        self.file = file

    def write(self, x):
        # Avoid print() second call (useless \n)
        # if len(x.rstrip()) > 0:
        tqdm.write(x)
        # self.buf = buf.strip("\r\n\t ")

    def flush(self):
        return getattr(self.file, "flush", lambda: None)()


LOGGING_CONFIG_DEFAULTS = dict(
    version=1,
    disable_existing_loggers=False,
    loggers={
        "sachima.root": {"level": "INFO", "handlers": ["file", "console"]},
        "sachima.error": {
            "level": "INFO",
            "handlers": ["error_console"],
            "propagate": True,
            "qualname": "sachima.error",
        },
    },
    handlers={
        "file": {
            "class": "logging.handlers.RotatingFileHandler",
            "formatter": "generic",
            "filename": conf.get("LOG_DIR") + "/sachima.log",  # noqa
            "maxBytes": 1024000,
            "backupCount": 10,
            # "stream": sys.stdout,
        },
        # "console1": {
        #     "class": "logging.StreamHandler",
        #     "formatter": "generic",
        #     "level": "DEBUG",
        # },
        "console": {
            "class": "logging.StreamHandler",
            "formatter": "generic",
            "level": "DEBUG",
            "stream": DummyTqdmFile(sys.stderr),
        },
        "error_console": {
            "class": "logging.StreamHandler",
            "formatter": "generic",
            "stream": sys.stderr,
        },
    },
    formatters={
        "generic": {
            "format": "%(asctime)s [%(process)d] [%(processName)s] [%(levelname)s] %(message)s",
            "datefmt": "[%Y-%m-%d %H:%M:%S %z]",
            "class": "logging.Formatter",
        }
    },
)

logging.config.dictConfig(LOGGING_CONFIG_DEFAULTS)
logger = logging.getLogger("sachima.root")
logger.setLevel(conf.get("LOG_LEVEL"))



import json as js
import os
import re

from sachima.log import logger


def drill(json_dict, path_desc):
    if not isinstance(json_dict, dict):
        raise TypeError("wrong type for first param should be dict")
    for element in path_desc:
        logger.debug("handle element：" + str(element))
        if len(json_dict) == 0:
            return ""
        # logger.debug(json_dict)
        if isinstance(element, str):
            # logger.debug("type(json_dict): " + str(type(json_dict)))
            if isinstance(json_dict, str):
                m = re.search(element, json_dict)
                if m is None:
                    return ""
                else:
                    return m.group()
            else:
                json_dict = json_dict.get(element, {})
                logger.debug("json_dict update: " + str(json_dict))
        elif isinstance(element, list):
            if len(element) == 1:  # 判断key是否存在
                # logger.debug('len(element) == 1')
                brk = False
                for o in json_dict:
                    if element[0] in o.keys():
                        json_dict = o[element[0]]
                        logger.debug(
                            "json_dict update: type: "
                            + str(type(json_dict))
                            + " len: "
                            + str(len(json_dict))
                        )
                        brk = True
                        break
                if brk is False:
                    return ""  # 如果没有匹配到，返回空字符串
            elif len(element) == 2:  # value值和给定值匹配
                # logger.debug('len(element) == 2')
                brk = False
                for o in json_dict:
                    if o[element[0]] == element[1]:
                        json_dict = o
                        logger.debug(
                            "json_dict update: type: "
                            + str(type(json_dict))
                            + " len: "
                            + str(len(json_dict))
                        )
                        brk = True
                        break
                if brk is False:
                    return ""  # 如果没有匹配到，返回空字符串
                # return('')  # 如果没有匹配到，返回空字符串
            elif len(element) == 3:
                # logger.debug('len(element) == 3')
                # ['detail','CONTAINS','3个月身份证关联家庭地址数'],['detail','REGEXP','\d$']
                # logger.debug("-------------------------"+str(element))
                if element[1] == "CONTAINS":
                    # logger.debug("-------------------------")
                    brk = False
                    for o in json_dict:
                        if element[2] in o[element[0]]:
                            json_dict = o
                            logger.debug(
                                "json_dict update: type: "
                                + str(type(json_dict))
                                + " len: "
                                + str(len(json_dict))
                            )
                            brk = True
                            break
                    if brk is False:
                        return ""  # 如果没有匹配到，返回空字符串
                if element[1] == "GETLIST" and isinstance(element[2], int):
                    json_dict = json_dict.get(element[0], {})[element[2]]
                    logger.debug(
                        "json_dict update: type: "
                        + str(type(json_dict))
                        + " len: "
                        + str(len(json_dict))
                    )
                elif element[1] == "GETLIST" and isinstance(element[2], str):
                    json_dict = json_dict.get(element[0], {})
                    brk = False
                    for o in json_dict:
                        if isinstance(o, str):
                            # print(element[2]+'-000000000000000000000'+o)
                            m = re.search(element[2], o)
                            if m is None:
                                continue
                            else:
                                json_dict = m.group()
                                logger.debug(
                                    "json_dict update: type: "
                                    + str(type(json_dict))
                                    + " len: "
                                    + str(len(json_dict))
                                )
                                brk = True
                                break
                    if brk is False:
                        return ""  # 如果没有匹配到，返回空字符串
    return json_dict


def flatten(f, conf_dict):
    """
    input json str
    flatten json and return sql
    """
    if isinstance(f, str):
        json = js.loads(f)
    elif isinstance(f, dict):
        json = f
    else:
        json = js.load(f)

    row = {}
    for key in conf_dict:
        logger.debug("handle column: " + key)
        row[key] = drill(json, conf_dict[key])

    # 转换成str
    for key in row:
        row[key] = str(row[key])

    return row


if __name__ == "__main__":
    pass



import sys
# import thread
import webbrowser
import time

import SimpleHTTPServer
import urllib
import SocketServer

PORT = 80

class Proxy(SimpleHTTPServer.SimpleHTTPRequestHandler):
    def do_GET(self):
        self.copyfile(urllib.urlopen(self.path), self.wfile)


def start_server():
    httpd = BaseHTTPServer.HTTPServer(('127.0.0.1', 3600), SimpleHTTPServer.SimpleHTTPRequestHandler)
    httpd.serve_forever()

thread.start_new_thread(start_server,())
url = 'http://127.0.0.1:3600'
webbrowser.open_new(url)

while True:
    try:
        time.sleep(1)
    except KeyboardInterrupt:
        sys.exit(0)


from nameko.rpc import rpc, RpcProxy
import importlib
import os
import numpy as np
import pandas as pd
import json

from sachima.log import logger
from sachima.cache import RedisClient
from sachima.tools import Tools

from sachima import conf

# ERROR_GRP_TOKEN = conf.get("SNS_DINGDING_ERROR_GRP_TOKEN")
# INFO_GRP_TOKEN = conf.get("SNS_DINGDING_INFO_GRP_TOKEN")
CACHE_LIST = conf.get("CACHE_LIST", [])


def data_wrapper(data):
    """
    data: dict
    return: json str
    return json str to api for frontend \n
    for example:
        antd
    """
    # data["data"]
    # data["filters"]
    # data["link"]
    # print(data)
    if not data:
        return {
            "columns": [{"title": "提示信息", "dataIndex": "提示信息", "key": "提示信息"}],
            "dataSource": [{"提示信息": "服务器数据出现错误请联系管理员"}],
        }

    # logger.info("data into json str...")
    res = {}
    df = data["data"][0]
    filters = data["filters"]
    links = data.get("links", {})
    vis = data.get("vis")

    if isinstance(df, pd.DataFrame):
        res["controls"] = [f.to_json(df) for f in filters]
        res["columns"] = [
            {"title": x, "dataIndex": x, "key": x, "render": {"action": links[x]},}
            if x in links
            else {"title": x, "dataIndex": x, "key": x}
            for x in df.columns
        ]
        # for colname in links:
        #     res["columns"][colname]["render"] = {
        #         "action": "http://www.baidu.com/s?wd="
        #     }
        # df = df.applymap(str)
        res["dataSource"] = json.loads(
            df.to_json(
                orient="records", date_format="iso", date_unit="s", force_ascii=False,
            )
        )
        res["vis"] = vis
        # logger.debug("res dataSource lens: " + str(len(res["dataSource"])))
        return res
    else:
        raise TypeError("your handler should return pd.DataFrame")


class Data(object):
    name = "data"

    @rpc
    def get_report(self, params):
        """
        todo: 加入缓存功能
        1.缓存出故障通知但不影响应用返回 done
        2.让调用方可以强制更新缓存  done
        3.批量刷新缓存（时间间隔，表格配置，不配的默认不刷新）
        4.晚间定时全部刷新
        5.设置最大内存，缓存淘汰策略
        """
        logger.info("=" * 50)
        logger.info(params)
        logger.info("=" * 50)
        need_fallback = False
        force_flash = params.get("isForce", 0)  # 0 使用缓存  1 不使用缓存
        params.pop("isForce", None)  # 把请求中控制是否强制刷新的flag去掉
        req = json.dumps(params)
        req_md5 = Tools.get_md5_value(req)

        try:
            conn = RedisClient().conn
            ret = conn.hget("sachima_results", req_md5)
            conn.zincrby("sachima_board", 1, req_md5)  # 监控访问次数

        except Exception as ex:
            # print(TypeError, ValueError)
            logger.info(
                "cache exception {} raised, fallbacking..... will not use cache".format(
                    str(ex)
                )
            )
            need_fallback = True  # 缓存出现异常，需要降级回退处理

        # 不需要降级并且有返回值并且前端没有要求强制刷新并且命中缓存
        if (
            params.get("name") in CACHE_LIST
            and not need_fallback
            and not force_flash
            and ret
        ):
            # cache hit
            logger.debug("cache hit")
            return json.loads(ret)
        else:
            # cache miss
            logger.debug("cache miss")
            logger.debug("call service: " + params.get("name"))
            # importlib.invalidate_caches()
            m = importlib.import_module(params.get("name"))
            m = importlib.reload(m)
            res = data_wrapper(m.main(params))

            if not need_fallback and params.get("name") in CACHE_LIST:
                conn.hset("sachima_req", req_md5, req)
                conn.hset("sachima_results", req_md5, json.dumps(res))  # 更新缓存
        return res

    @rpc
    def writecache(self, name, value, conn=RedisClient().conn):
        logger.debug("writecache to: " + name)
        conn.rpush(
            name, json.dumps(value, indent=2, ensure_ascii=False).encode("utf-8"),
        )

    @rpc
    def popcache(self, name, conn=RedisClient().conn):
        logger.debug("lpopcache: " + name)
        return conn.lpop(name)



from flask import Flask, jsonify, request
import json

from sachima.rpc import Data
from sachima import conf

CONFIG = conf.get("BROKER")

app = Flask(__name__)


def sachima_rpc_reports(req, local=True):
    print("*"*80)
    print(req.json)
    d = Data()
    res = d.get_report(req.json)

    # TODO: is Waffle call me twice? improve it
    if req.json.get("dataonly") == "True":
        return {
            "dataSource": res.get("dataSource", []),
            "columns": res.get("columns", []),
        }
    else:
        return res


@app.route("/reports", methods=["POST"])
def sachima():
    # print(request.json)
    return jsonify(sachima_rpc_reports(request))




"""
if you need to connect Oracle  you shoud install cx_Oracle

mac instanclient install
http://www.oracle.com/technetwork/topics/intel-macsoft-096467.html?ssSourceSiteId=otncn

mkdir ~/lib
ln -s ~/instantclient_12_2/libclntsh.dylib ~/lib/
ln -s ~/instantclient_12_2/libclntsh.dylib.12.1 ~/lib/


https://stackoverflow.com/questions/51170169/clean-up-database-connection-with-sqlalchemy-in-pandas
engine = create_engine('...', poolclass=NullPool)
engine.dispose()
"""
# https://github.com/cloudera/impyla
from impala.dbapi import connect as imp_connect
from pymysql import connect as my_connect
from pymysql import cursors as cursors

# from impala.util import as_pandas
from sqlalchemy import create_engine, MetaData, pool
from sqlalchemy import Table, Column, Date, Integer, String, ForeignKey
from sachima import conf

# ENGINE_ORACLE = create_engine(
#         "oracle+cx_oracle://xxx:xxxx@xxxxx:1521/POS",
#         connect_args={'encoding': 'utf8', 'nencoding': 'utf8'})


class db(object):
    @property
    def ENGINE_SUPERSET(self):
        return create_engine("sqlite:///{}".format(conf.get("DB_SUPERSET")), echo=True)

    @property
    def ENGINE_MYSQL_h(self):
        return create_engine(
            "mysql+pymysql://{}:{}@{}/{}".format(
                conf.get("MYSQL_H_USER"),
                conf.get("MYSQL_H_PASS"),
                conf.get("MYSQL_H_IP"),
                conf.get("MYSQL_H_DB"),
            ),
            connect_args={"charset": conf.get("MYSQL_H_CHARSET")},
            poolclass=pool.NullPool,
        )

    @property
    def ENGINE_MYSQL_d(self):
        return create_engine(
            "mysql+pymysql://{}:{}@{}/{}".format(
                conf.get("MYSQL_D_USER"),
                conf.get("MYSQL_D_PASS"),
                conf.get("MYSQL_D_IP"),
                conf.get("MYSQL_D_DB"),
            ),
            connect_args={"charset": conf.get("MYSQL_D_CHARSET")},
            poolclass=pool.NullPool,
        )

    @property
    def ENGINE_IMPALA_DW(self):
        def _impala_conn():
            return imp_connect(
                host=conf.get("IMPALA_IP"),
                port=conf.get("IMPALA_PORT"),
                database=conf.get("IMPALA_DB"),
                timeout=conf.get("IMPALA_TIMEOUT"),
                #      use_ssl=True,
                #      ca_cert='some_pem',
                #      user='cloudera',
                #      password='cloudera',
                auth_mechanism=conf.get("IMPALA_AUTH_MECHANISM"),
                #      kerberos_service_name='hive'
            )

        return create_engine(
            "impala://", creator=_impala_conn, echo=False, poolclass=pool.NullPool,
        )

    @property
    def ENGINE_MARIADB_dw(self):
        return create_engine(
            "mysql+pymysql://{}:{}@{}/{}".format(
                conf.get("MARIADB_USER"),
                conf.get("MARIADB_PASS"),
                conf.get("MARIADB_IP"),
                conf.get("MARIADB_DB"),
            ),
            connect_args={"charset": conf.get("MARIADB_CHARSET")},
            poolclass=pool.NullPool,
        )

    # ---------------------------------
    # conn
    # ---------------------------------

    @property
    def CONN_IMPALA_DW(self):
        return imp_connect(
            host=conf.get("IMPALA_IP"),
            port=conf.get("IMPALA_PORT"),
            database=conf.get("IMPALA_DB"),
            auth_mechanism=conf.get("IMPALA_AUTH_MECHANISM"),
        )

    @property
    def CONN_MYSQL_h(self):
        return my_connect(
            host=conf.get("MYSQL_H_IP"),
            user=conf.get("MYSQL_H_USER"),
            password=conf.get("MYSQL_H_PASS"),
            port=conf.get("MYSQL_H_PORT"),
            database=conf.get("MYSQL_H_DB"),
            charset=conf.get("MYSQL_H_CHARSET"),
            use_unicode=True,
            cursorclass=cursors.SSDictCursor,
        )

    @property
    def CONN_MYSQL_d(self):
        return my_connect(
            host=conf.get("MYSQL_D_IP"),
            user=conf.get("MYSQL_D_USER"),
            password=conf.get("MYSQL_D_PASS"),
            port=conf.get("MYSQL_D_PORT"),
            database=conf.get("MYSQL_D_DB"),
            charset=conf.get("MYSQL_D_CHARSET"),
            cursorclass=cursors.SSDictCursor,
        )

    # ---------------------------------
    # meta
    # ---------------------------------

    @property
    def MetaData_MYSQL_h(self):
        return MetaData(self.ENGINE_MYSQL_h)

    @property
    def MetaData_MYSQL_d(self):
        return MetaData(self.ENGINE_MYSQL_d)

    @property
    def MetaData_IMPALA_DW(self):
        return MetaData(self.ENGINE_IMPALA_DW)



from sachima.filter_enum import FilterEnum as _
from sachima.params import Filter

def append_select_filter(mainrunner_json_result, filters):
    '''
    add filters to main_runner results \n
    usage: append_select_filter(return_res_from_main,[{"name":"行纬度","option":[]},{"name":"列纬度","option":[]}] ) 
    '''
    for f in filters:
        mainrunner_json_result["filters"].append(
           Filter(
                f.get("name"),
                setter=(
                    _.TYPE.ITEMSELECT,
                    _.PROPS.ALLOWCLEAR.TRUE,
                    _.PROPS.MODE.MULTIPLE,
                    {"option": f.get("option")},
                ),
            ), 
        )



# -*- coding: utf-8 -*-
"""
@author: mk
"""


def app_style_to_excel(dfname,
                       sheetname,
                       range='A1:R',
                       autofilter=0,
                       style='Table Style Light 9'):
    columns = [{'header': c} for c in dfname.columns]
    options = {
        'style': 'Table Style Light 9',
        'header_row': True,
        'banded_columns': False,
        'autofilter': 0,
        'columns': columns
    }
    # Add a table to the worksheet.
    sheetname.add_table(range + str(len(dfname) + 1), options)



"""
title: s_sachima_funnel_example
"""
from sachima.main_runner import run
from sachima.filter_enum import FilterEnum as _
from sachima.params import Filter
from sachima.wrappers import send, timer

from services.db_connects import db
import datetime as dt

db = db()


@timer
def main(api_params={}):
    PARAM_IN = {
        "model": [("example_data/sankey.csv", "csv")],
        "handler": ["s_sankey"],
        "params": {},
        "filters": [],
        "vis": {
            "type": "sankey",
            "title": "Sachima Sankey Graph example",
            "params": {"layoutIterations": 0}, # default sequence 
        },
    }

    return run(PARAM_IN, api_params)


if __name__ == "__main__":
    main()




import hashlib
import os
import re
import sys
import uuid
from calendar import monthrange, weekday
from datetime import date, datetime, timedelta

import numpy as np
import pandas as pd

from sachima.tools import Tools
from sachima import conf

BASE_FILES_PATH = os.path.join(conf.get("PROJ_DIR"), "data")
OUT_PATH = BASE_FILES_PATH
CONF_PATH = os.path.join(BASE_FILES_PATH, "alertconf.csv")

BATCH_DATE = datetime.strptime(
    datetime.strftime(datetime.today(), "%Y-%m-%d"), "%Y-%m-%d"
)
TIME_LEN = 60
NEED_CHART = False

panel_data = {}


def handle_conf(conf):
    """
    Handle config file by
    1. Remove special whitespace
    2. Drop duplicate rules
    3. Add ruleid as first column

    ps: This function will modify the conf file inplace
    """
    # Open conf file
    ori = pd.read_csv(conf, index_col=None, sep=",")

    # Remove special whitespace
    ori.replace("\xa0", " ", inplace=True, regex=True)
    ori.replace("\s+", " ", inplace=True, regex=True)

    # Column names
    col_name = ori.columns.tolist()

    # Generate md5 value for each row and put it in the first column as ruleid
    ori["ruleid"] = (
        ori[col_name]
        .drop(["ruleid"], axis=1)
        .apply(
            lambda x: Tools.get_md5_value("-".join(str(value) for value in x)), axis=1,
        )
    )
    ori.set_index(["ruleid"], inplace=True)
    ori.drop_duplicates(inplace=True)

    # Save modified conf file
    ori.to_csv(conf)


def date_cut(x):
    """
    date_cut converts the incoming string date into datetime.datetime date type after processing
    example:
    2017-9 -> 2017-09-30 00:00:00
    2017-09-03 -> 2017-09-03 00:00:00
    20170701-20170801 -> 2017-08-01 00:00:00
    """
    if type(x) == str and len(x) <= 7:
        temp = x.split("-")
        iyear, imonth = int(temp[0]), int(temp[1])
        # str(monthrange(iyear, imonth)[1] the last day of the month
        sample = "-".join([x, str(monthrange(iyear, imonth)[1])])
        return datetime.strptime(sample, "%Y-%m-%d")
    elif type(x) == str and len(x) > 7 and len(x) <= 10:  # 2017-08-01   20170801
        for fmt in ("%Y-%m-%d", "%Y%m%d", "%Y/%m/%d"):
            try:
                return datetime.strptime(x, fmt)
            except ValueError:
                pass
        raise ValueError("Could not convert date")
    elif type(x) == str and len(x) > 10:  # 20170701-20170801
        return datetime.strptime(x.split("-")[1], "%Y%m%d")
    else:
        return x


def clean(groupdata):
    """
    Cut out incomplete data at the current time point
    """
    if type(groupdata) != pd.core.frame.DataFrame:
        raise ("DataFrame must be passed in")
    groupdata["stattime"] = groupdata.index.map(date_cut)
    return groupdata


def get_last_time(groupdata):
    """
    Use the system date to determine if there is excess time

    The expected incoming data set is each grouped data set, a dataframe after a combination of dimensions
    Then judge the different time types according to the length Weekly interval 20160829-20160904 Month 2017-12

    """
    t1 = BATCH_DATE + timedelta(days=-1)

    try:
        the_end_date_put_in = groupdata.stattime[-1]
    except:
        print(
            "There is no data after clean, the data becomes empty after being cleaned in the latest period"
        )
        return ""

    # if the_end_date_put_in != t1:
    #     print(the_end_date_put_in)
    #     print(t1)
    #     print("*********The final time is not t-1, not calculated")
    #     return ""
    # else:
    return groupdata.index[-1]


def rulerun(ruleid, table, index, column, rule, param, url, f):
    """
    example: 
        ruleid=ff27a4ff5ea11f3c319670037f32316f
        table=risk.table1 
        index=dim1 dim2 td_stat_date
        column=column1 column2
        rule=R00003_multicondi
        param=< > 0.8 0.1
        url=www.google.com
        f=open("filename", "w")
    """
    df = panel_data[table]
    index = index.split(" ")  # example: d_dim1 d_dim2 td_stat_date

    # get begin time and end time
    dt_end = max(df[index[-1]])
    if len(df[index[-1]].drop_duplicates(keep="first", inplace=False)) > TIME_LEN:
        dt_begin = (
            df[index[-1]]
            .drop_duplicates(keep="first", inplace=False)
            .sort_values()
            .iloc[-TIME_LEN]
        )
    else:
        dt_begin = min(df[index[-1]])

    # get begin time to end time range data
    df = df[(df[index[-1]] >= dt_begin) & (df[index[-1]] <= dt_end)]

    df = df.set_index(index).sort_index()
    ruletype = rule.split("_")[0]  # example： R00001_xxxxx
    eval(ruletype)(ruleid, df, table, column, param, index, url, f)


def R00001(ruleid, df, table, column, param, index, url, f):
    """
    R00001_{L}-order difference value does not exceed {N} daily moving window {M} upper and lower standard deviations
    """
    s_param = param.split(" ")
    # Parameters are extracted in order, and when configured in the configuration file, they are also in order.
    diff_deep = int(s_param[0])
    # Moving window length
    window_value = int(s_param[1])
    # How many times the standard deviation is considered an outlier
    std_times = float(s_param[2])
    # Field name of upper standard deviation
    std_top = "diff_" + str(std_times) + "times_mstd" + str(window_value)
    # The field name of the lower standard deviation
    std_bottom = "diff_-" + str(std_times) + "times_mstd" + str(window_value)

    if len(index) > 1:
        dim_str = str(index[0:-1])  # Except the last one as the dimension
        time_str = str(index[-1])  # The last one is the time x axis
        dimention = df.groupby(index[0:-1]).count().index  # Combination of dimensions
    else:
        dim_str = "_"  # In order to run on win, special characters such as asterisk or less than greater than sign cannot be used
        time_str = str(index[0])
        dimention = ["_"]  # There is no other dimension except time

    for groupset in dimention:
        print(
            "R00001"
            + ruleid
            + "------------Processing group---"
            + table
            # + filename
            + str(index)
            + "--------"
            + str(groupset)
            + "---------------Processing group----------------"
        )
        dt_end = ""
        dd = None

        if groupset == "_":
            dd = clean(df)
            dt_end = get_last_time(dd)
        else:
            dd = clean(df.loc[groupset])
            dt_end = get_last_time(dd)

        if type(dt_end) == str and dt_end == "":
            print("----in continue------")
            continue

        obj = dd[:dt_end][[column]]

        if obj.size <= 7:
            print("Time series is less than 7. exit")
            continue

        # Difference The difference between the previous time and the previous time. The need_diff parameter controls whether the difference is needed, and the diff_deep parameter controls the order of the diff
        obj["diff"] = obj[
            column
        ]  # Increase the diff field equal to the original value when initialized
        i = diff_deep  # diff_deep order difference
        while i > 0:
            obj["diff"] = obj["diff"].diff()
            i = i - 1

        # Moving average Moving average window_value day average
        obj["diff_ma" + str(window_value)] = (
            obj["diff"].rolling(window=window_value, center=False).mean()
        )
        # Moving std moving standard deviation std_times times standard deviation
        obj[std_top] = (
            std_times * obj["diff"].rolling(window=window_value, center=False).std()
        )
        obj[std_bottom] = (
            -std_times * obj["diff"].rolling(window=window_value, center=False).std()
        )

        # Determine whether the indicator is within the range Increase the upward or downward fluctuation
        conditions = [
            (obj["diff"] > obj[std_top]) | (obj["diff"] < obj[std_bottom]),
            1 == 1,
        ]
        conditios_up = [(obj["diff"] > obj[std_top]), 1 == 1]
        conditios_down = [(obj["diff"] < obj[std_bottom]), 1 == 1]
        choices = [True, False]  # True means an alert is triggered
        obj["res"] = np.select(conditions, choices, default=False)
        obj["up"] = np.select(conditios_up, choices, default=False)
        obj["down"] = np.select(conditios_down, choices, default=False)

        if (
            obj.loc[obj.index[-1]]["res"] == False
        ):  # The last point must be an early warning point, otherwise no warning is output
            print(
                "The last point must be an early warning point, otherwise no warning is output"
            )
            continue

        # Only record the last warning point
        last_dt = obj[obj["res"] == True].index[-1]
        vv = obj.loc[last_dt][column]  # Trigger warning value
        sBD = "Upward Fluctuation"
        if obj.loc[last_dt]["up"] == True:
            sBD = "Upward Fluctuation"
        elif obj.loc[last_dt]["up"] == False:
            sBD = "Downward Fluctuate"

        alerttext = (
            Tools.special_char_remove(dim_str)
            + ","
            + Tools.special_char_remove(str(groupset))
            + ","
            + table
            + ",["
            + column
            + "],"
            + sBD
            + " Abnormal,"
            + param
        )
        data_file = os.path.join("..", table + ".xlsx")

        f.write(
            str(uuid.uuid1())
            + ","
            + ruleid
            + ",R00001,[Volatility warning],"
            + Tools.time_dim_to_str(last_dt)
            + ",["
            + time_str
            + "],"
            + alerttext
            + ","
            + Tools.special_char_remove(str(vv))
            + ","
            + ""
            + str(url)
            + ""
            + ',=HYPERLINK("'
            + data_file
            + '")'
            + ',=HYPERLINK("'
            + str(url)
            + '")'
            + "\n"
        )
        f.flush()


def R00005(ruleid, df, table, column, param, index, url, f):
    """
    R00005_Absolute value{>=<}{T}
    """
    # fill None value to Zero
    df.fillna(0, inplace=True)
    # params example: "< 100"
    # so we split it with whitespace then we got the operation and threshold value
    params = param.split(" ")
    oper = params[0]  # >=<
    value = float(params[1])  # Threshold

    # If the length of index is one, it means only the time dimension, if it is greater than 1, take the dimension index[0:-1] other than the last time dimension as groupby
    if len(index) > 1:
        dim_str = str(index[0:-1])  # Other dimensions
        time_str = str(index[-1])  # Time dimension
        temp = (
            df.groupby(index[0:-1]).count().index
        )  # List of other dimensions, group processing
    else:
        time_str = str(index[0])
        dim_str = "_"
        temp = ["_"]  # There is no other dimension except time

    for groupset in temp:
        print(
            "R00005"
            + "--------ruleid:"
            + ruleid
            + "------------Processing group---"
            + table
            + "@"
            + str(index)
            + "--------"
            + str(groupset)
            + "----------"
        )
        dt_end = ""
        dd = None

        if groupset == "_":
            dd = clean(df)
            dt_end = get_last_time(dd)
        else:
            dd = clean(df.loc[groupset])
            dt_end = get_last_time(dd)

        if type(dt_end) == str and dt_end == "":
            continue

        obj = dd[:dt_end][[column]]

        obj["thresholdvalue"] = value
        obj["-thresholdvalue"] = -1 * value
        if oper == ">":
            conditions = [abs(obj[column]) > obj["thresholdvalue"], 1 == 1]
            print("conditions: ", conditions)
            print("thresholdvalue", obj["thresholdvalue"])
            oper_str = " greater than "
        elif oper == "<":
            conditions = [abs(obj[column]) < obj["thresholdvalue"], 1 == 1]
            oper_str = " less than "
        elif oper == "=":
            conditions = [abs(obj[column]) == obj["thresholdvalue"], 1 == 1]
            oper_str = " is equal to "

        choices = [True, False]
        obj["res"] = np.select(conditions, choices, default=False)

        alerttext = (
            Tools.special_char_remove(dim_str)
            + ","
            + Tools.special_char_remove(str(groupset))
            + ","
            + table
            + ",["
            + column
            + "],Absolute value"
            + oper_str
            + "Threshold,"
            + str(value)
        )

        # The last point must be an early warning point, otherwise no warning will be output and cannot be used is must == there will be bugs
        if obj.loc[obj.index[-1]]["res"] == False:
            continue

        last_dt = obj[obj["res"] == True].index[-1]
        vv = obj.loc[last_dt][column]
        data_file = os.path.join("..", table + ".xlsx")
        f.write(
            str(uuid.uuid1())
            + ","
            + ruleid
            + ",R00005,[High limit warning],"
            + Tools.time_dim_to_str(last_dt)
            + ",["
            + time_str
            + "],"
            + alerttext
            + ","
            + Tools.special_char_remove(str(vv))
            + ","
            + ""
            + str(url)
            + ""
            + ',=HYPERLINK("'
            + data_file
            + '")'
            + ',=HYPERLINK("'
            + str(url)
            + '")'
            + "\n"
        )
        f.flush()


def R00003(ruleid, df, table, column, param, index, url, f):
    """
    R00003 multi-condition rule matching,
    """

    column_lists = column.split(" ")
    param_lists = param.split(" ")
    l = len(column_lists)
    oper_lists = param_lists[:l]
    value_lists = param_lists[l:]
    value_lists = [Tools.maybe_float(v) for v in value_lists]

    if len(column_lists) == len(oper_lists) == len(value_lists):
        pass
    else:
        print(
            "Please enter the matching field parameters and values, separated by a space, this rule will not be executed："
            + ruleid
        )
        return 0

    if (
        len(index) > 1
    ):  # If the length of index is one, it means there is only time dimension. If it is greater than 1, take the dimension idx[0:-1] other than the last time dimension as groupby
        dim_str = str(index[0:-1])  # Other dimensions
        time_str = str(index[-1])  # Time dimension
        dimention = (
            df.groupby(index[0:-1]).count().index
        )  # List of other dimensions, group processing
    else:
        time_str = str(index[0])
        dim_str = "_"
        dimention = ["_"]  # There is no other dimension except time

    for groupset in dimention:
        print(
            "R00003--------ruleid:"
            + ruleid
            + "------------Processing group----"
            + table
            + str(index)
            + "-------"
            + str(groupset)
            + "----------"
        )
        dt_end = ""
        dd = None

        if groupset == "_":
            dd = clean(df)
            dt_end = get_last_time(dd)
        else:
            dd = clean(df.loc[groupset])
            dt_end = get_last_time(dd)

        if type(dt_end) == str and dt_end == "":
            continue

        obj = dd[:dt_end][list(set(column_lists))]
        c_markers_lists = []
        for c, v, o in zip(column_lists, value_lists, oper_lists):
            keyname = (
                c + o + str(v)
            )  # The field operator and the value form a unique keyname to form a new column
            # print(keyname)
            obj[keyname + "_thresholdvalue"] = v
            obj[keyname + "markers"] = Tools.get_truth(
                obj[c], o, obj[keyname + "_thresholdvalue"]
            )
            c_markers_lists.append(keyname + "markers")

        obj["res"] = obj[c_markers_lists].all(
            axis=1
        )  # All conditions must be met to trigger an alert

        alerttext = (
            Tools.special_char_remove(dim_str)
            + ","
            + Tools.special_char_remove(str(groupset))
            + ","
            + table
            + ","
            + Tools.special_char_remove(str(column_lists))
            + ","
            + Tools.special_char_remove(str(oper_lists))
            + ","
            + Tools.special_char_remove(str(value_lists))
        )

        if (
            obj.loc[obj.index[-1]]["res"] == False
        ):  # The last point must be an early warning point, otherwise no warning is output
            continue

        last_dt = obj[obj["res"] == True].index[-1]
        vv = list(obj.loc[last_dt][column_lists])
        f.write(
            str(uuid.uuid1())
            + ","
            + ruleid
            + ",R00003,[Multi-condition warning],"
            + Tools.time_dim_to_str(last_dt)
            + ",["
            + time_str
            + "],"
            + alerttext
            + ","
            + Tools.special_char_remove(str(vv))
            + ","
            + str(url)
            + ',=HYPERLINK("'
            + str(url)
            + '")'
            + "\n"
        )
        f.flush()


def get(days=1, conf_path=CONF_PATH):
    """
    get alert out put
    """
    for d in range(0, days):
        BATCH_DATE = datetime.strptime(
            datetime.strftime(datetime.today(), "%Y-%m-%d"), "%Y-%m-%d"
        ) + timedelta(days=-d)

        # read alert conf file
        conf = pd.read_csv(conf_path, index_col=None, sep=",")
        # you can config active==N in config file to disable the alert rule
        conf = conf[conf["active"] == "Y"]

        # the path store results
        last_path = os.path.join(
            OUT_PATH, "lastest_alerts" + datetime.strftime(BATCH_DATE, "%Y%m%d"),
        )

        if not os.path.exists(last_path):
            os.makedirs(last_path)

        csv_file_name = os.path.join(
            last_path, "alertlist" + datetime.strftime(BATCH_DATE, "%Y%m%d") + ".csv",
        )

        f_last_alerts = open(csv_file_name, "w", encoding="utf-8")
        f_last_alerts.write(
            "alarmid,ruleid,trigger_rule,rule_type,date,date_type,dimension_name,dimension_value,table,field,description,parameter,value,graph,data,url"
            + "\n"
        )
        f_last_alerts.flush()

        for index, row in conf.iterrows():
            rulerun(
                row["ruleid"],
                row["table"],
                row["index"],
                row["column"],
                row["rule"],
                row["param"],
                row["url"],
                f_last_alerts,
            )

        f_last_alerts.close()

        return pd.read_csv(csv_file_name, encoding="utf-8")


def add(name, df):
    """
    add df to global panel_data 
    name: path.filename
    df: input pandas Dataframe
    """
    panel_data[name] = df




import pandas as pd
from sachima import conf
import os
import importlib
from sachima.params import set_sql_params
from sachima.log import logger
from sachima.wrappers import timer

# from tqdm import tqdm
import io
import time
import logging
import sys
import threading
import itertools


@timer
def sql(sql, datatype):
    """
    return DataFrame
    """
    logger.info("running sql {}".format(sql))
    return pd.read_sql(sql, datatype)


class Data:
    def __init__(self, dataname, datatype, params, prefunc):
        """
        dataname: sql filename
        datatype: db engine  or filetype in str
        """
        self.source = dataname
        self.type = datatype
        title = dataname + " | " + str(datatype)
        logger.info("=" * 12 + " " + title + " " + "=" * 12)
        if datatype in ("xls", "xlsx"):
            self.data = pd.read_excel(os.path.join(conf.get("PROJ_DIR"), dataname))
        elif datatype in ("csv", "txt"):
            self.data = pd.read_csv(os.path.join(conf.get("PROJ_DIR"), dataname))
        elif datatype in ("api",):
            api_cls = importlib.import_module("services." + dataname, package="..")
            api = api_cls.Api()
            self.data = api.data
        elif datatype in ("json"):
            self.data = pd.read_json(os.path.join(conf.get("PROJ_DIR"), dataname))
        else:
            # read sql file from ./sqls
            str_sql = open(
                os.path.join(conf.get("PROJ_DIR"), "sqls", dataname), encoding="utf-8",
            ).read()
            sql = str_sql
            # pre process before sql loaded
            if prefunc:
                sql = prefunc(set_sql_params(str_sql, params), params)
            else:
                sql = set_sql_params(str_sql, params)

            self.data = _get_df(sql, datatype, dataname)


def animate(dataname, log):
    t = threading.currentThread()
    elapsed_time = 0
    for c in itertools.cycle(["|", "/", "-", "\\"]):
        if getattr(t, "Done", True):
            break
        sys.stdout.write("\r<{}> running {} {} ms\r".format(dataname, c, elapsed_time))
        sys.stdout.flush()
        time.sleep(0.1)
        elapsed_time += 100

    # log("<{}> parsing elapsed time: {} ms".format(dataname, elapsed_time))
    log("\r")
    # sys.stdout.write("\r\n Done!     ")


def _get_df(sql, datatype, dataname):
    animate_thread = threading.Thread(target=animate, args=(dataname, logger.info))
    animate_thread.daemon = True
    animate_thread.Done = False
    animate_thread.start()

    try:
        start = time.time()
        df = pd.read_sql(sql, datatype)
        consumed_time = time.time() - start

        logger.info("<{}> time: {} secs".format(dataname, consumed_time))
    except Exception as e:
        raise e
    finally:
        animate_thread.Done = True  # no matter how break the animate loop
    # logger.info("<{}> start loading data... ".format(dataname))

    # # df = pd.concat(first + [chunk for chunk in tqdm(chunks, total=200)])

    # logger.info(
    #     "<{}> loading data elapsed time: {} seconds".format(
    #         dataname, loading_data_elapsed_time
    #     )
    # )
    return df

